"""Tests for `imitation.algorithms.adversarial.*` algorithms."""

import contextlib
import os
from typing import Any, Mapping, Optional, Type, Union

import numpy as np
import pytest
import seals  # noqa: F401
import stable_baselines3
import torch as th
from stable_baselines3.common import policies
from torch.utils import data as th_data

from imitation.algorithms.adversarial import airl, common, gail
from imitation.data import rollout, types
from imitation.rewards import reward_nets
from imitation.util import logger, util

ALGORITHM_KWARGS = {
    "airl-ppo": {
        "algorithm_cls": airl.AIRL,
        "model_class": stable_baselines3.PPO,
        "policy_class": policies.ActorCriticPolicy,
    },
    "gail-ppo": {
        "algorithm_cls": gail.GAIL,
        "model_class": stable_baselines3.PPO,
        "policy_class": policies.ActorCriticPolicy,
    },
    "gail-dqn": {
        "algorithm_cls": gail.GAIL,
        "model_class": stable_baselines3.DQN,
        "policy_class": stable_baselines3.dqn.MlpPolicy,
    },
}
IN_CODECOV = "COV_CORE_CONFIG" in os.environ
# Disable SubprocVecEnv tests for code coverage test since
# multiprocessing support is flaky in py.test --cov
PARALLEL = [False] if IN_CODECOV else [True, False]
ENV_NAMES = ["FrozenLake-v1", "CartPole-v1", "Pendulum-v1"]
EXPERT_BATCH_SIZES = [1, 128]


@pytest.fixture(params=ALGORITHM_KWARGS.values(), ids=list(ALGORITHM_KWARGS.keys()))
def _algorithm_kwargs(request):
    """Auto-parametrizes `_rl_algorithm_cls` for the `trainer` fixture."""
    return dict(request.param)


@pytest.fixture
def expert_transitions(cartpole_expert_trajectories):
    return rollout.flatten_trajectories(cartpole_expert_trajectories)


@contextlib.contextmanager
def make_trainer(
    algorithm_kwargs: Mapping[str, Any],
    tmpdir: str,
    expert_transitions: types.Transitions,
    rng: np.random.Generator,
    expert_batch_size: int = 1,
    env_name: str = "seals/CartPole-v0",
    num_envs: int = 1,
    parallel: bool = False,
    convert_dataset: bool = False,
    **kwargs: Any,
):
    expert_data: Union[th_data.DataLoader, th_data.Dataset]
    if convert_dataset:
        expert_data = th_data.DataLoader(
            expert_transitions,
            batch_size=expert_batch_size,
            collate_fn=types.transitions_collate_fn,
            shuffle=True,
            drop_last=True,
        )
    else:
        expert_data = expert_transitions

    venv = util.make_vec_env(
        env_name,
        n_envs=num_envs,
        parallel=parallel,
        rng=rng,
    )
    model_cls = algorithm_kwargs["model_class"]
    gen_algo = model_cls(algorithm_kwargs["policy_class"], venv)
    reward_net_cls: Type[reward_nets.RewardNet] = reward_nets.BasicRewardNet
    if algorithm_kwargs["algorithm_cls"] == airl.AIRL:
        reward_net_cls = reward_nets.BasicShapedRewardNet
    reward_net = reward_net_cls(venv.observation_space, venv.action_space)
    custom_logger = logger.configure(tmpdir, ["tensorboard", "stdout"])

    trainer = algorithm_kwargs["algorithm_cls"](
        venv=venv,
        demonstrations=expert_data,
        demo_batch_size=expert_batch_size,
        gen_algo=gen_algo,
        reward_net=reward_net,
        log_dir=tmpdir,
        custom_logger=custom_logger,
        **kwargs,
    )

    try:
        yield trainer
    finally:
        venv.close()


def test_airl_fail_fast(custom_logger, tmpdir, rng):
    venv = util.make_vec_env(
        "seals/CartPole-v0",
        n_envs=1,
        parallel=False,
        rng=rng,
    )

    gen_algo = stable_baselines3.DQN(stable_baselines3.dqn.MlpPolicy, venv)
    small_data = rollout.generate_transitions(
        gen_algo,
        venv,
        n_timesteps=20,
        rng=rng,
    )
    reward_net = reward_nets.BasicShapedRewardNet(
        observation_space=venv.observation_space,
        action_space=venv.action_space,
    )

    with pytest.raises(TypeError, match="AIRL needs a stochastic policy.*"):
        airl.AIRL(
            venv=venv,
            demonstrations=small_data,
            demo_batch_size=20,
            gen_algo=gen_algo,
            reward_net=reward_net,
            log_dir=tmpdir,
            custom_logger=custom_logger,
        )


@pytest.fixture(params=ALGORITHM_KWARGS.values(), ids=list(ALGORITHM_KWARGS.keys()))
def trainer(request, tmpdir, expert_transitions, rng):
    with make_trainer(
        request.param,
        tmpdir,
        expert_transitions,
        rng,
    ) as trainer:
        yield trainer


def test_train_disc_no_samples_error(trainer: common.AdversarialTrainer):
    with pytest.raises(RuntimeError, match="No generator samples"):
        trainer.train_disc()


def test_train_disc_unequal_expert_gen_samples_error(
    trainer: common.AdversarialTrainer,
    expert_transitions: types.Transitions,
):
    """Test that train_disc raises error when n_gen != n_expert samples."""
    if len(expert_transitions) < 2:  # pragma: no cover
        raise ValueError("Test assumes at least 2 samples.")
    expert_samples = types.dataclass_quick_asdict(expert_transitions)
    gen_samples = types.dataclass_quick_asdict(expert_transitions[:-1])
    with pytest.raises(ValueError, match="n_expert"):
        trainer.train_disc(expert_samples=expert_samples, gen_samples=gen_samples)


@pytest.fixture(params=PARALLEL)
def _parallel(request):
    """Auto-parametrizes `_parallel`."""
    return request.param


@pytest.fixture(params=[False, True])
def _convert_dataset(request):
    """Auto-parametrizes `_parallel`."""
    return request.param


@pytest.fixture(params=[1, 128])
def _expert_batch_size(request):
    """Auto-parameterizes `_expert_batch_size`."""
    return request.param


@pytest.fixture
def trainer_parametrized(
    _algorithm_kwargs,
    _parallel,
    _convert_dataset,
    _expert_batch_size,
    tmpdir,
    expert_transitions,
    rng,
):
    with make_trainer(
        _algorithm_kwargs,
        tmpdir,
        expert_transitions,
        rng=rng,
        parallel=_parallel,
        convert_dataset=_convert_dataset,
        expert_batch_size=_expert_batch_size,
    ) as trainer:
        yield trainer


def test_train_disc_step_no_crash(
    trainer_parametrized,
    _expert_batch_size,
    rng,
):
    transitions = rollout.generate_transitions(
        trainer_parametrized.gen_algo,
        trainer_parametrized.venv,
        n_timesteps=_expert_batch_size,
        truncate=True,
        rng=rng,
    )
    trainer_parametrized.train_disc(
        gen_samples=types.dataclass_quick_asdict(transitions),
    )


def test_train_gen_train_disc_no_crash(
    trainer_parametrized: common.AdversarialTrainer,
    n_updates: int = 2,
) -> None:
    trainer_parametrized.train_gen(n_updates * trainer_parametrized.gen_train_timesteps)
    trainer_parametrized.train_disc()


@pytest.fixture
def trainer_batch_sizes(
    _algorithm_kwargs,
    _expert_batch_size,
    tmpdir,
    expert_transitions,
    rng,
):
    with make_trainer(
        _algorithm_kwargs,
        tmpdir,
        expert_transitions,
        expert_batch_size=_expert_batch_size,
        rng=rng,
    ) as trainer:
        yield trainer


def test_train_disc_improve_D(
    trainer_batch_sizes,
    tmpdir,
    expert_transitions,
    _expert_batch_size,
    rng,
    n_steps=3,
):
    expert_samples = expert_transitions[:_expert_batch_size]
    expert_samples = types.dataclass_quick_asdict(expert_samples)
    gen_samples = rollout.generate_transitions(
        trainer_batch_sizes.gen_algo,
        trainer_batch_sizes.venv_train,
        n_timesteps=_expert_batch_size,
        truncate=True,
        rng=rng,
    )
    gen_samples = types.dataclass_quick_asdict(gen_samples)
    init_stats = final_stats = None
    for _ in range(n_steps):
        final_stats = trainer_batch_sizes.train_disc(
            gen_samples=gen_samples,
            expert_samples=expert_samples,
        )
        if init_stats is None:
            init_stats = final_stats
    assert final_stats["disc_loss"] < init_stats["disc_loss"]


def test_gradient_accumulation(
    _algorithm_kwargs,
    tmpdir,
    expert_transitions,
    rng,
    cartpole_venv,
):
    batch_size = 6
    minibatch_size = 3

    expert_samples = expert_transitions[:batch_size]
    expert_samples = types.dataclass_quick_asdict(expert_samples)

    # Sample actions randomly to produce mock generator data
    gen_samples_trans = rollout.generate_transitions(
        policy=None,
        venv=cartpole_venv,
        n_timesteps=batch_size,
        truncate=True,
        rng=rng,
    )
    gen_samples = types.dataclass_quick_asdict(gen_samples_trans)

    seed = rng.integers(2**32)

    def trainer_ctx(minibatch_size: Optional[int] = None):
        rng = np.random.default_rng(seed)
        th.manual_seed(seed)
        return make_trainer(
            _algorithm_kwargs,
            tmpdir,
            expert_transitions,
            rng,
            batch_size,
            demo_minibatch_size=minibatch_size,
        )

    with trainer_ctx() as trainer1, trainer_ctx(minibatch_size) as trainer2:
        for step in range(8):
            print("Step", step)
            for trainer in (trainer1, trainer2):
                trainer.train_disc(
                    gen_samples=gen_samples,
                    expert_samples=expert_samples,
                )

            # Note: due to numerical instability, the models are
            # bound to diverge at some point, but should be stable
            # over the short time frame we test over; however, it is
            # theoretically possible that with very unlucky seeding,
            # this could fail.
            params = zip(
                trainer1._reward_net.parameters(),
                trainer2._reward_net.parameters(),
            )
            atol = (1 + step) * 2e-4
            rtol = (1 + step) * 1e-5
            for p1, p2 in params:
                th.testing.assert_close(p1, p2, atol=atol, rtol=rtol)


@pytest.fixture(params=ENV_NAMES)
def _env_name(request):
    """Auto-parameterizes `_env_name`."""
    return request.param


@pytest.fixture
def trainer_diverse_env(
    _algorithm_kwargs,
    _env_name,
    tmpdir,
    expert_transitions,
    rng,
):
    if _algorithm_kwargs["model_class"] == stable_baselines3.DQN:
        pytest.skip("DQN does not support all environments.")
    with make_trainer(
        _algorithm_kwargs,
        tmpdir,
        expert_transitions,
        rng=rng,
        env_name=_env_name,
    ) as trainer:
        yield trainer


@pytest.mark.parametrize("n_timesteps", [2, 4, 10])
def test_logits_expert_is_high_log_policy_act_prob(
    trainer_diverse_env: common.AdversarialTrainer,
    n_timesteps: int,
    rng,
):
    """Smoke test calling `logits_expert_is_high` on `AdversarialTrainer`.

    For AIRL, also checks that the function raises
    error on `log_policy_act_prob=None`.

    Args:
        trainer_diverse_env: The trainer to test.
        n_timesteps: The number of timesteps of rollouts to collect.
        rng: The random state to use.
    """
    trans = rollout.generate_transitions(
        policy=None,
        venv=trainer_diverse_env.venv,
        n_timesteps=n_timesteps,
        rng=rng,
    )

    obs, acts, next_obs, dones = trainer_diverse_env.reward_train.preprocess(
        types.assert_not_dictobs(trans.obs),
        trans.acts,
        types.assert_not_dictobs(trans.next_obs),
        trans.dones,
    )
    log_act_prob_non_none = np.log(0.1 + 0.9 * np.random.rand(n_timesteps))
    log_act_prob_non_none_th = th.as_tensor(log_act_prob_non_none).to(obs.device)

    for log_act_prob in [None, log_act_prob_non_none_th]:
        maybe_error_ctx: contextlib.AbstractContextManager
        if isinstance(trainer_diverse_env, airl.AIRL) and log_act_prob is None:
            maybe_error_ctx = pytest.raises(TypeError, match="Non-None.*required.*")
        else:
            maybe_error_ctx = contextlib.nullcontext()

        with maybe_error_ctx:
            trainer_diverse_env.logits_expert_is_high(
                obs,
                acts,
                next_obs,
                dones,
                log_act_prob,
            )


@pytest.mark.parametrize("n_samples", [0, 1, 10, 40])
def test_compute_train_stats(n_samples):
    disc_logits_expert_is_high = th.from_numpy(
        np.random.standard_normal([n_samples]) * 10,
    )
    labels_expert_is_one = th.from_numpy(np.random.randint(2, size=[n_samples]))
    disc_loss = th.tensor(np.random.random() * 10)
    stats = common.compute_train_stats(
        disc_logits_expert_is_high,
        labels_expert_is_one,
        disc_loss,
    )
    for k, v in stats.items():
        assert isinstance(k, str)
        assert isinstance(v, float)


@pytest.mark.skipif(not th.cuda.is_available(), reason="requires GPU")
def test_regression_gail_with_sac(
    pendulum_expert_trajectories,
    pendulum_venv,
):  # pragma: no cover
    """GAIL with a SAC learner on GPU used to crash when training (see #655).

    This is a minimal test to reproduce it.

    Args:
        pendulum_expert_trajectories: expert trajectories for Pendulum env.
        pendulum_venv: the Pendulum environment.
    """
    learner = stable_baselines3.SAC(
        env=pendulum_venv,
        policy=stable_baselines3.sac.policies.SACPolicy,
    )
    reward_net = reward_nets.BasicRewardNet(
        pendulum_venv.observation_space,
        pendulum_venv.action_space,
    )
    gail_trainer = gail.GAIL(
        demonstrations=pendulum_expert_trajectories,
        demo_batch_size=1024,
        venv=pendulum_venv,
        gen_algo=learner,
        reward_net=reward_net,
    )
    gail_trainer.train(8)



"""Compute the probability that one algorithm improved over another."""
import argparse
import dataclasses
import pathlib
import warnings
from typing import Dict, List, Optional

import numpy as np
from rliable import library as rly
from rliable import metrics

from imitation.util.sacred_file_parsing import SacredRun, group_runs_by_algo_and_env


def make_score_matrix_from_runs_by_env(
    runs_by_env: Dict[str, List[SacredRun]],
    envs: Optional[List[str]] = None,
) -> np.ndarray:
    """Creates a matrix of scores from the runs for each environment.

    Note: when the number of samples for each environment is not equal, the samples
    will be truncated to the minimum sample count.

    Args:
        runs_by_env: A dictionary mapping environment names to lists of runs.
        envs: The environments to sample from. If None, all environments are used.

    Returns:
        A matrix of scores of shape (n_samples, n_envs).
    """
    if envs is None:
        envs = list(runs_by_env.keys())

    sample_counts_by_env = {env: len(runs_by_env[env]) for env in envs}

    min_sample_count = min(sample_counts_by_env.values())
    if not all(
        sample_counts_by_env[env] == sample_counts_by_env[envs[0]] for env in envs
    ):
        warnings.warn(
            f"The runs for the environments have different sample counts "
            f"{sample_counts_by_env}. "
            f"This is not supported by the probability of improvement. Therefore, "
            f"samples will be truncated to the minimum sample count of"
            f" {min_sample_count}",
        )

    return np.asarray(
        [
            [
                run["result"]["imit_stats"]["monitor_return_mean"]
                for run in runs_by_env[env][:min_sample_count]
            ]
            for env in envs
        ],
    ).T


@dataclasses.dataclass
class ProbabilityOfImprovementResult:
    """The result of the probability of improvement computation.

    Attributes:
        probability_of_improvement: The probability of improvement.
        confidence_interval: The confidence interval of the probability of improvement.
        samples_per_env: The number of samples per environment after truncation.
        baseline_samples_per_env: The number of baseline samples per environment after
            truncation.
    """

    probability_of_improvement: float
    confidence_interval: np.ndarray
    samples_per_env: int
    baseline_samples_per_env: int


def compute_probability_of_improvement(
    runs_by_env: Dict[str, List[SacredRun]],
    baseline_runs_by_env: Dict[str, List[SacredRun]],
    reps: int,
):
    """Computes the probability of improvement of the runs over the baseline runs.

    Args:
        runs_by_env: A dictionary mapping environment names to lists of runs.
        baseline_runs_by_env: A dictionary mapping environment names to lists of runs.
        reps: The number of bootstrap repetitions to use to compute the confidence
            interval.

    Returns:
        A ProbabilityOfImprovementResult object that contains the probability with its
        confidence interval and some information about the number of samples that has
        effectively been used.
    """
    envs = set(runs_by_env.keys())
    baseline_envs = set(baseline_runs_by_env.keys())
    comparison_envs = sorted(envs & baseline_envs)

    # We need to arrange the scores in a matrix of shape (n_samples, n_envs).
    # When we do not have the same number of samples for each environment,
    # the samples are truncated to the minimum sample count.
    run_scores = make_score_matrix_from_runs_by_env(runs_by_env, comparison_envs)
    assert run_scores.shape[0] >= 1
    assert run_scores.shape[1] == len(comparison_envs)
    baseline_run_scores = make_score_matrix_from_runs_by_env(
        baseline_runs_by_env,
        comparison_envs,
    )
    assert baseline_run_scores.shape[0] >= 1
    assert baseline_run_scores.shape[1] == len(comparison_envs)

    samples_per_env = run_scores.shape[0]
    baseline_samples_per_env = baseline_run_scores.shape[0]

    probabs, error_intervals = rly.get_interval_estimates(
        {"baseline_vs_new": (baseline_run_scores, run_scores)},
        metrics.probability_of_improvement,
        reps=reps,
    )
    probability_of_improvement = probabs["baseline_vs_new"]
    confidence_interval = np.squeeze(error_intervals["baseline_vs_new"])

    return ProbabilityOfImprovementResult(
        probability_of_improvement,
        confidence_interval,
        samples_per_env,
        baseline_samples_per_env,
    )


def main():
    parser = argparse.ArgumentParser(
        description="Compute the probability of improvement of one algorithm over "
        "another by comparing its runs to some baseline runs.",
    )
    parser.add_argument(
        "runs_dir",
        type=pathlib.Path,
        help="A directory containing sacred output directories with runs for the "
        "algorithm that should have improved.",
    )
    parser.add_argument(
        "baseline_runs_dir",
        nargs="?",
        default=None,
        type=pathlib.Path,
        help="A directory containing sacred output directories with runs for the "
        "algorithm to compare against. If not specified, the runs_dir will be "
        "used. In that case make sure to specify the --baseline-algo option."
        "Otherwise you just compare an algorithm to itself.",
    )
    parser.add_argument(
        "--algo",
        type=str,
        help="The algorithm to compare. Only needs to be specified when the runs_dir "
        "contains runs from more than one algorithm.",
    )
    parser.add_argument(
        "--baseline-algo",
        type=str,
        help="The algorithm to compare against. By default, either the single "
        "algorithm in the baseline_runs_dir is used, or the same algorithm as "
        "specified by the --algo option is used.",
    )
    parser.add_argument(
        "--bootstrap-reps",
        type=int,
        default=2000,
        help="The number of bootstrap repetitions to use to compute the stratified "
        "confidence intervals.",
    )

    args = parser.parse_args()

    if args.baseline_runs_dir is None:
        args.baseline_runs_dir = args.runs_dir

    runs_by_algo_and_env = group_runs_by_algo_and_env(
        args.runs_dir,
        only_completed_runs=True,
    )
    baseline_runs_by_algo_and_env = group_runs_by_algo_and_env(
        args.baseline_runs_dir,
        only_completed_runs=True,
    )

    algos_in_run_dir = sorted(runs_by_algo_and_env.keys())
    algos_in_baseline_dir = sorted(baseline_runs_by_algo_and_env.keys())

    if len(algos_in_run_dir) == 0:
        raise ValueError(f"The run directory [{args.runs_dir}] contains no runs.")

    if len(algos_in_baseline_dir) == 0:
        raise ValueError(
            f"The baseline run directory [{args.baseline_runs_dir}] "
            f"contains no runs.",
        )

    if args.algo is None:
        if len(algos_in_run_dir) == 1:
            args.algo = algos_in_run_dir[0]
        else:
            raise ValueError(
                f"The run directory [{args.runs_dir}] contains runs for the "
                f"algorithms [{', '.join(algos_in_run_dir)}]. Please use the --algo "
                f"option to specify which algorithms runs to compare.",
            )

    if args.baseline_algo is None:
        if len(algos_in_baseline_dir) == 1:
            # If the baseline algo is not specified and the baseline run dir contains
            # only one algo, we assume that the user wants to compare to that algo.
            args.baseline_algo = algos_in_baseline_dir[0]
        elif args.algo in algos_in_baseline_dir:
            # If the baseline algo is not specified and there is more than one option
            # to pick from, and the algo is in the baseline run dir, we assume that
            # the user wants to compare to the same algo between the two run dirs.
            args.baseline_algo = args.algo
        else:
            raise ValueError(
                f"The baseline run directory [{args.baseline_runs_dir}] contains "
                f"runs for the algorithms [{', '.join(algos_in_baseline_dir)}]. "
                f"Please use the --baseline-algo option specify which one to "
                f"compare to.",
            )

    if args.algo not in algos_in_run_dir:
        raise ValueError(
            f"The run directory [{args.runs_dir}] contains runs for the algorithms"
            f" [{', '.join(algos_in_run_dir)}]. You specified [{args.algo}], for which "
            f"no runs can be found in the run directory",
        )

    if args.baseline_algo not in algos_in_baseline_dir:
        raise ValueError(
            f"The baseline run directory [{args.baseline_runs_dir}] contains runs "
            f"for the algorithms [{', '.join(algos_in_baseline_dir)}]. "
            f"You specified [{args.baseline_algo}], for which no runs can be found"
            f" in the baseline run directory",
        )

    if (args.algo == args.baseline_algo) and (args.runs_dir == args.baseline_runs_dir):
        warnings.warn(
            "You are comparing two equal sets of runs. "
            "This is probably not what you want.",
        )

    envs = set(runs_by_algo_and_env[args.algo].keys())
    baseline_envs = set(baseline_runs_by_algo_and_env[args.baseline_algo].keys())

    comparison_envs = envs & baseline_envs

    if len(comparison_envs) == 0:
        raise ValueError(
            f"The baseline runs are for the environments "
            f"[{', '.join(baseline_envs)}], while the runs are for the "
            f"environments [{', '.join(envs)}]. "
            f"There is no overlap in the environments of the two run sets, so no "
            f"comparison can be made",
        )

    if envs != baseline_envs:
        warnings.warn(
            f"The baseline runs are for the environments "
            f"[{', '.join(baseline_envs)}], "
            f"while the runs are for the environments [{', '.join(envs)}]. "
            f"The comparison will only be made for the environments "
            f"[{', '.join(comparison_envs)}].",
        )

    probability_of_improvement_result = compute_probability_of_improvement(
        runs_by_env=runs_by_algo_and_env[args.algo],
        baseline_runs_by_env=baseline_runs_by_algo_and_env[args.baseline_algo],
        reps=args.bootstrap_reps,
    )

    algo_str = args.algo
    baseline_algo_str = args.baseline_algo
    if args.algo == args.baseline_algo:
        # Only print the run directories if we have to disambiguate the two algorithms
        # When we have to different algorithms we can assume the user knows which
        # runs_dir and baseline_runs_dir they are using.
        algo_str += f" ({args.runs_dir})"
        baseline_algo_str += f" ({args.baseline_runs_dir})"

    print(
        f"Comparison based on {probability_of_improvement_result.n_samples} samples "
        f"per environment for {algo_str} and "
        f"{probability_of_improvement_result.n_baseline_samples} samples per "
        f"environment for {baseline_algo_str}.",
    )
    print(f"Samples taken in {', '.join(comparison_envs)}")
    print()
    print(f"Probability of improvement of {algo_str} over {baseline_algo_str}:")
    print(
        f"{probability_of_improvement_result.probability_of_improvement:.3f} "
        f"({probability_of_improvement_result.confidence_interval[0]:.3f}, "
        f"{probability_of_improvement_result.confidence_interval[1]:.3f}, "
        f"reps={args.bootstrap_reps:,})",
    )


if __name__ == "__main__":
    main()



"""Soft Q Imitation Learning (SQIL) (https://arxiv.org/abs/1905.11108).

Trains a policy via DQN-style Q-learning,
replacing half the buffer with expert demonstrations and adjusting the rewards.
"""

from typing import Any, Dict, List, Optional, Type, Union

import numpy as np
import torch as th
from gymnasium import spaces
from stable_baselines3 import dqn
from stable_baselines3.common import (
    buffers,
    off_policy_algorithm,
    policies,
    type_aliases,
    vec_env,
)

from imitation.algorithms import base as algo_base
from imitation.data import rollout, types
from imitation.util import logger, util


class SQIL(algo_base.DemonstrationAlgorithm[types.Transitions]):
    """Soft Q Imitation Learning (SQIL).

    Trains a policy via DQN-style Q-learning,
    replacing half the buffer with expert demonstrations and adjusting the rewards.
    """

    expert_buffer: buffers.ReplayBuffer

    def __init__(
        self,
        *,
        venv: vec_env.VecEnv,
        demonstrations: Optional[algo_base.AnyTransitions],
        policy: Union[str, Type[policies.BasePolicy]],
        custom_logger: Optional[logger.HierarchicalLogger] = None,
        rl_algo_class: Type[off_policy_algorithm.OffPolicyAlgorithm] = dqn.DQN,
        rl_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """Builds SQIL.

        Args:
            venv: The vectorized environment to train on.
            demonstrations: Demonstrations to use for training.
            policy: The policy model to use (SB3).
            custom_logger: Where to log to; if None (default), creates a new logger.
            rl_algo_class: Off-policy RL algorithm to use.
            rl_kwargs: Keyword arguments to pass to the RL algorithm constructor.

        Raises:
            ValueError: if `dqn_kwargs` includes a key
                `replay_buffer_class` or `replay_buffer_kwargs`.
        """
        self.venv = venv

        if rl_kwargs is None:
            rl_kwargs = {}
        # SOMEDAY(adam): we could support users specifying their own replay buffer
        # if we made SQILReplayBuffer a more flexible wrapper. Does not seem worth
        # the added complexity until we have a concrete use case, however.
        if "replay_buffer_class" in rl_kwargs:
            raise ValueError(
                "SQIL uses a custom replay buffer: "
                "'replay_buffer_class' not allowed.",
            )
        if "replay_buffer_kwargs" in rl_kwargs:
            raise ValueError(
                "SQIL uses a custom replay buffer: "
                "'replay_buffer_kwargs' not allowed.",
            )

        self.rl_algo = rl_algo_class(
            policy=policy,
            env=venv,
            replay_buffer_class=SQILReplayBuffer,
            replay_buffer_kwargs={"demonstrations": demonstrations},
            **rl_kwargs,
        )

        super().__init__(demonstrations=demonstrations, custom_logger=custom_logger)

    def set_demonstrations(self, demonstrations: algo_base.AnyTransitions) -> None:
        assert isinstance(self.rl_algo.replay_buffer, SQILReplayBuffer)
        self.rl_algo.replay_buffer.set_demonstrations(demonstrations)

    def train(self, *, total_timesteps: int, tb_log_name: str = "SQIL", **kwargs: Any):
        self.rl_algo.learn(
            total_timesteps=total_timesteps,
            tb_log_name=tb_log_name,
            **kwargs,
        )

    @property
    def policy(self) -> policies.BasePolicy:
        assert isinstance(self.rl_algo.policy, policies.BasePolicy)
        return self.rl_algo.policy


class SQILReplayBuffer(buffers.ReplayBuffer):
    """A replay buffer that injects 50% expert demonstrations when sampling.

    This buffer is fundamentally the same as ReplayBuffer,
    but it includes an expert demonstration internal buffer.
    When sampling a batch of data, it will be 50/50 expert and collected data.

    It can be used in off-policy algorithms like DQN/SAC/TD3.

    Here it is used as part of SQIL, where it is used to train a DQN.
    """

    def __init__(
        self,
        buffer_size: int,
        observation_space: spaces.Space,
        action_space: spaces.Space,
        demonstrations: algo_base.AnyTransitions,
        device: Union[th.device, str] = "auto",
        n_envs: int = 1,
        optimize_memory_usage: bool = False,
    ):
        """Create a SQILReplayBuffer instance.

        Args:
            buffer_size: Max number of elements in the buffer
            observation_space: Observation space
            action_space: Action space
            demonstrations: Expert demonstrations.
            device: PyTorch device.
            n_envs: Number of parallel environments. Defaults to 1.
            optimize_memory_usage: Enable a memory efficient variant
                of the replay buffer which reduces by almost a factor two
                the memory used, at a cost of more complexity.
        """
        super().__init__(
            buffer_size=buffer_size,
            observation_space=observation_space,
            action_space=action_space,
            device=device,
            n_envs=n_envs,
            optimize_memory_usage=optimize_memory_usage,
            handle_timeout_termination=False,
        )

        self.expert_buffer = buffers.ReplayBuffer(
            buffer_size=0,
            observation_space=observation_space,
            action_space=action_space,
        )
        self.set_demonstrations(demonstrations)

    def set_demonstrations(
        self,
        demonstrations: algo_base.AnyTransitions,
    ) -> None:
        """Set the expert demonstrations to be injected when sampling from the buffer.

        Args:
            demonstrations (algo_base.AnyTransitions): Expert demonstrations.

        Raises:
            NotImplementedError: If `demonstrations` is not a transitions object
                or a list of trajectories.
        """
        # If demonstrations is a list of trajectories,
        # flatten it into a list of transitions
        if not isinstance(demonstrations, types.Transitions):
            (
                item,
                demonstrations,
            ) = util.get_first_iter_element(  # type: ignore[assignment]
                demonstrations,  # type: ignore[assignment]
            )
            if isinstance(item, types.Trajectory):
                demonstrations = rollout.flatten_trajectories(
                    demonstrations,  # type: ignore[arg-type]
                )

        if not isinstance(demonstrations, types.Transitions):
            raise NotImplementedError(
                f"Unsupported demonstrations type: {demonstrations}",
            )

        n_samples = len(demonstrations)
        self.expert_buffer = buffers.ReplayBuffer(
            buffer_size=n_samples,
            observation_space=self.observation_space,
            action_space=self.action_space,
            handle_timeout_termination=False,
        )

        for transition in demonstrations:
            self.expert_buffer.add(
                obs=transition["obs"],
                next_obs=transition["next_obs"],
                action=transition["acts"],
                done=transition["dones"],
                reward=np.array(1.0),
                infos=[{}],
            )

    def add(
        self,
        obs: np.ndarray,
        next_obs: np.ndarray,
        action: np.ndarray,
        reward: np.ndarray,
        done: np.ndarray,
        infos: List[Dict[str, Any]],
    ) -> None:
        super().add(
            obs=obs,
            next_obs=next_obs,
            action=action,
            reward=np.array(0.0),
            done=done,
            infos=infos,
        )

    def sample(
        self,
        batch_size: int,
        env: Optional[vec_env.VecNormalize] = None,
    ) -> buffers.ReplayBufferSamples:
        """Sample a batch of data.

        Half of the batch will be from expert transitions,
        and the other half will be from the learner transitions.

        Args:
            batch_size: Number of elements to sample in total
            env: associated gym VecEnv to normalize the observations/rewards
                when sampling

        Returns:
            A mix of transitions from the expert and from the learner.
        """
        new_sample_size, expert_sample_size = util.split_in_half(batch_size)
        new_sample = super().sample(new_sample_size, env)
        expert_sample = self.expert_buffer.sample(expert_sample_size, env)

        return type_aliases.ReplayBufferSamples(
            *(
                th.cat((getattr(new_sample, name), getattr(expert_sample, name)))
                for name in new_sample._fields
            ),
        )



"""Tests for `imitation.data.buffer`."""

import gymnasium as gym
import numpy as np
import pytest

from imitation.data import types
from imitation.data.buffer import Buffer, ReplayBuffer


def _fill_chunk(start, chunk_len, sample_shape, dtype=float):
    fill_vals = np.arange(start, start + chunk_len, dtype=dtype)
    fill_vals = np.reshape(fill_vals, (-1,) + (1,) * len(sample_shape))
    chunk = np.tile(fill_vals, (1,) + sample_shape)
    return chunk


def _get_fill_from_chunk(chunk):
    chunk_len, *sample_shape = chunk.shape
    sample_size = max(1, np.prod(sample_shape))
    return chunk.flatten()[::sample_size]


def _check_bound(end, capacity, samples, offset=0):
    start = max(0, end - capacity)
    assert np.all(start + offset <= samples), "samples violate lower bound"
    assert np.all(samples <= end + offset), "samples violate upper bound"


@pytest.mark.parametrize("capacity", [10, 30, 60])
@pytest.mark.parametrize("chunk_len", [1, 2, 4, 9])
@pytest.mark.parametrize("sample_shape", [(), (1, 2), (5, 4, 4)])
def test_buffer(capacity, chunk_len, sample_shape) -> None:
    """Tests `buffer.Buffer` by creating a buffer, inserting data and checking samples.

    Builds a Buffer with the provided `capacity` and inserts `capacity * 3`
    samples into the buffer in chunks of shape `(chunk_len,) + sample_shape`.

    We always insert chunks with consecutive integers.

    The test checks that:

        - `len(buffer)` increases until we reach capacity.
        - `buffer._idx` loops between 0 and `capacity - 1`.
        - After every insertion, samples are in the expected range, verifying
      FIFO insertion.
        - Mutating the inserted chunk doesn't mutate the buffer.

    Args:
        capacity: The capacity of the buffer to create.
        chunk_len: The number of chunks to insert in one go.
        sample_shape: The shape of the data to insert.
    """
    buf = Buffer(
        capacity,
        sample_shapes={"a": sample_shape, "b": sample_shape},
        dtypes={"a": np.dtype(float), "b": np.dtype(float)},
    )

    to_insert = 3 * capacity
    for i in range(0, to_insert, chunk_len):
        assert buf.size() == min(i, capacity)
        assert buf._idx == i % capacity
        chunk_a = _fill_chunk(i, chunk_len, sample_shape)
        chunk_b = _fill_chunk(i + to_insert, chunk_len, sample_shape)
        buf.store({"a": chunk_a, "b": chunk_b})
        samples = buf.sample(100)
        assert set(samples.keys()) == {"a", "b"}, samples.keys()
        _check_bound(i + chunk_len, capacity, samples["a"])
        _check_bound(i + chunk_len + to_insert, capacity, samples["b"])
        assert np.all(samples["b"] - samples["a"] == to_insert)

        # Confirm that buffer is not mutable from inserted sample.
        chunk_a[:] = np.nan
        chunk_b[:] = np.nan
        assert not np.any(np.isnan(buf._arrays["a"]))
        assert not np.any(np.isnan(buf._arrays["b"]))


@pytest.mark.parametrize("capacity", [30, 60])
@pytest.mark.parametrize("chunk_len", [1, 4, 9])
@pytest.mark.parametrize("obs_shape", [(), (1, 2)])
@pytest.mark.parametrize("act_shape", [(), (5, 4, 4)])
@pytest.mark.parametrize("dtype", [int, np.float32])
def test_replay_buffer(capacity, chunk_len, obs_shape, act_shape, dtype):
    """Tests `ReplayBuffer` by creating a buffer, inserting data and checking samples.

    Inserts `capacity * 3` observation-action-observation samples into the buffer in
    chunks of length `chunk_len`.

    All chunks are of the appropriate observation or action shape, and contain
    the value fill_val.

    Tests that:

        - len(buffer)` increases until we reach capacity.
        - `buffer._idx` loops between 0 and `capacity - 1`.
        - After every insertion, samples only contain samples from
          expected range.

    Args:
        capacity: The capacity of the `ReplayBuffer`.
        chunk_len: The length of each chunk to insert.
        obs_shape: Shape of observations.
        act_shape: Shape of actions.
        dtype: dtype used for observations and actions.

    """
    buf = ReplayBuffer(
        capacity,
        obs_shape=obs_shape,
        act_shape=act_shape,
        obs_dtype=dtype,
        act_dtype=dtype,
    )

    for i in range(0, capacity * 3, chunk_len):
        assert buf.size() == min(i, capacity)
        assert buf._buffer._idx == i % capacity

        dones = np.arange(i, i + chunk_len, dtype=np.int32) % 2
        dones = dones.astype(bool)
        infos = _fill_chunk(9 * capacity + i, chunk_len, (), dtype=dtype)
        infos = np.array([{"a": val} for val in infos])
        batch = types.Transitions(
            obs=_fill_chunk(i, chunk_len, obs_shape, dtype=dtype),
            next_obs=_fill_chunk(3 * capacity + i, chunk_len, obs_shape, dtype=dtype),
            acts=_fill_chunk(6 * capacity + i, chunk_len, act_shape, dtype=dtype),
            dones=dones,
            infos=infos,
        )
        buf.store(batch)

        # Are samples right shape?
        sample = buf.sample(100)
        info_vals = np.array([info["a"] for info in sample.infos])

        # dictobs not supported for buffers, or by current code in
        # this test file (eg `_get_fill_from_chunk`)
        obs = types.assert_not_dictobs(sample.obs)
        next_obs = types.assert_not_dictobs(sample.next_obs)

        assert obs.shape == next_obs.shape == (100,) + obs_shape
        assert sample.acts.shape == (100,) + act_shape
        assert sample.dones.shape == (100,)
        assert info_vals.shape == (100,)

        # Are samples right data type?
        assert obs.dtype == dtype
        assert sample.acts.dtype == dtype
        assert next_obs.dtype == dtype
        assert info_vals.dtype == dtype
        assert sample.dones.dtype == bool
        assert sample.infos.dtype == object

        # Are samples in range?
        _check_bound(i + chunk_len, capacity, obs)
        _check_bound(i + chunk_len, capacity, next_obs, 3 * capacity)
        _check_bound(i + chunk_len, capacity, sample.acts, 6 * capacity)
        _check_bound(i + chunk_len, capacity, info_vals, 9 * capacity)

        # Are samples in-order?
        obs_fill = _get_fill_from_chunk(obs)
        next_obs_fill = _get_fill_from_chunk(next_obs)
        act_fill = _get_fill_from_chunk(sample.acts)
        info_vals_fill = _get_fill_from_chunk(info_vals)

        assert np.all(next_obs_fill - obs_fill == 3 * capacity), "out of order"
        assert np.all(act_fill - next_obs_fill == 3 * capacity), "out of order"
        assert np.all(info_vals_fill - act_fill == 3 * capacity), "out of order"
        # Can't do much other than parity check for boolean values.
        # `samples.done` has the same parity as `obs_fill` by construction.
        assert np.all(obs_fill % 2 == sample.dones), "out of order"


@pytest.mark.parametrize("sample_shape", [(), (1,), (5, 2)])
def test_buffer_store_errors(sample_shape):
    capacity = 11
    dtype = np.float32

    def buf():
        return Buffer(capacity, {"k": sample_shape}, {"k": dtype})

    # Unexpected keys
    b = buf()
    with pytest.raises(ValueError):
        b.store({})
    chunk = np.ones((1,) + sample_shape)
    with pytest.raises(ValueError):
        b.store({"y": chunk})
    with pytest.raises(ValueError):
        b.store({"k": chunk, "y": chunk})

    # `data` is empty.
    b = buf()
    with pytest.raises(ValueError):
        b.store({"k": np.empty((0,) + sample_shape, dtype=dtype)})

    # `data` has too many samples.
    b = buf()
    with pytest.raises(ValueError):
        b.store({"k": np.empty((capacity + 1,) + sample_shape, dtype=dtype)})

    # `data` has the wrong sample shape.
    b = buf()
    with pytest.raises(ValueError):
        b.store({"k": np.empty((1, 3, 3, 3, 3), dtype=dtype)})


def test_buffer_sample_errors():
    b = Buffer(10, {"k": (2, 1)}, dtypes={"k": np.bool_})
    with pytest.raises(ValueError):
        b.sample(5)


def test_buffer_init_errors():
    with pytest.raises(KeyError, match=r"sample_shape and dtypes.*"):
        Buffer(10, dict(a=(2, 1), b=(3,)), dtypes=dict(a=np.float32, c=np.bool_))


def test_replay_buffer_init_errors():
    with pytest.raises(
        ValueError,
        match=r"Cannot specify both observation shape and also environment",
    ):
        ReplayBuffer(15, venv=gym.make("CartPole-v1"), obs_shape=(10, 10))
    with pytest.raises(ValueError, match=r"Shape or dtype missing.*"):
        ReplayBuffer(15, obs_shape=(10, 10), act_shape=(15,), obs_dtype=np.bool_)
    with pytest.raises(ValueError, match=r"Shape or dtype missing.*"):
        ReplayBuffer(15, obs_shape=(10, 10), obs_dtype=np.bool_, act_dtype=np.bool_)


def test_buffer_from_data():
    data = np.ndarray([50, 30], dtype=np.bool_)
    buf = Buffer.from_data({"k": data})
    assert buf._arrays["k"] is not data
    assert data.dtype == buf._arrays["k"].dtype
    assert np.array_equal(buf._arrays["k"], data)


def test_replay_buffer_from_data():
    obs = np.array([5, 2], dtype=int)
    acts = np.ones((2, 6), dtype=float)
    next_obs = np.array([7, 8], dtype=int)
    dones = np.array([True, False])
    infos = np.array([{}, {"a": "sdf"}])

    def _check_buf(buf):
        assert np.array_equal(buf._buffer._arrays["obs"], obs)
        assert np.array_equal(buf._buffer._arrays["next_obs"], next_obs)
        assert np.array_equal(buf._buffer._arrays["acts"], acts)
        assert np.array_equal(buf._buffer._arrays["infos"], infos)

    buf_std = ReplayBuffer.from_data(
        types.Transitions(
            obs=obs,
            acts=acts,
            next_obs=next_obs,
            dones=dones,
            infos=infos,
        ),
    )
    _check_buf(buf_std)

    rews = np.array([0.5, 1.0], dtype=float)
    buf_rew = ReplayBuffer.from_data(
        types.TransitionsWithRew(
            obs=obs,
            acts=acts,
            next_obs=next_obs,
            rews=rews,
            dones=dones,
            infos=infos,
        ),
    )
    _check_buf(buf_rew)



"""Tests for the preference comparisons reward learning implementation."""

import math
import re
from typing import Any, Sequence

import gymnasium as gym
import numpy as np
import pytest
import seals  # noqa: F401
import stable_baselines3
import torch as th
from gymnasium import spaces
from stable_baselines3.common import evaluation
from stable_baselines3.common.envs import FakeImageEnv
from stable_baselines3.common.vec_env import DummyVecEnv

import imitation.testing.reward_nets as testing_reward_nets
from imitation.algorithms import preference_comparisons
from imitation.data import types
from imitation.data.types import TrajectoryWithRew
from imitation.regularization import regularizers, updaters
from imitation.rewards import reward_nets
from imitation.testing import reward_improvement
from imitation.util import networks, util

UNCERTAINTY_ON = ["logit", "probability", "label"]


@pytest.fixture
def venv(rng):
    return util.make_vec_env(
        "seals/CartPole-v0",
        n_envs=1,
        rng=rng,
    )


@pytest.fixture(
    params=[
        reward_nets.BasicRewardNet,
        testing_reward_nets.make_ensemble,
        lambda *args: reward_nets.AddSTDRewardWrapper(
            testing_reward_nets.make_ensemble(*args),
        ),
    ],
)
def reward_net(request, venv):
    return request.param(venv.observation_space, venv.action_space)


@pytest.fixture
def agent(venv):
    return stable_baselines3.PPO(
        "MlpPolicy",
        venv,
        n_epochs=1,
        batch_size=2,
        n_steps=10,
    )


@pytest.fixture
def random_fragmenter(rng):
    return preference_comparisons.RandomFragmenter(
        rng=rng,
        warning_threshold=0,
    )


@pytest.fixture
def agent_trainer(agent, reward_net, venv, rng):
    return preference_comparisons.AgentTrainer(agent, reward_net, venv, rng)


def assert_info_arrs_equal(arr1, arr2):  # pragma: no cover
    def check_possibly_nested_dicts_equal(dict1, dict2):
        for key, val1 in dict1.items():
            val2 = dict2[key]
            if isinstance(val1, dict):
                check_possibly_nested_dicts_equal(val1, val2)
            else:
                assert np.array_equal(val1, val2)

    for item1, item2 in zip(arr1, arr2):
        assert isinstance(item1, dict)
        assert isinstance(item2, dict)
        check_possibly_nested_dicts_equal(item1, item2)


def _check_trajs_equal(
    trajs1: Sequence[types.TrajectoryWithRew],
    trajs2: Sequence[types.TrajectoryWithRew],
):
    assert len(trajs1) == len(trajs2)
    for traj1, traj2 in zip(trajs1, trajs2):
        assert np.array_equal(
            types.assert_not_dictobs(traj1.obs),
            types.assert_not_dictobs(traj2.obs),
        )
        assert np.array_equal(traj1.acts, traj2.acts)
        assert np.array_equal(traj1.rews, traj2.rews)
        assert traj1.infos is not None
        assert traj2.infos is not None
        assert_info_arrs_equal(traj1.infos, traj2.infos)
        assert traj1.terminal == traj2.terminal


def test_mismatched_spaces(venv, agent, rng):
    other_venv = util.make_vec_env(
        "seals/MountainCar-v0",
        n_envs=1,
        rng=rng,
    )
    bad_reward_net = reward_nets.BasicRewardNet(
        other_venv.observation_space,
        other_venv.action_space,
    )
    with pytest.raises(
        ValueError,
        match="spaces do not match",
    ):
        preference_comparisons.AgentTrainer(
            agent,
            bad_reward_net,
            venv,
            rng=rng,
        )


def test_trajectory_dataset_seeding(
    cartpole_expert_trajectories: Sequence[TrajectoryWithRew],
    num_samples: int = 400,
):
    dataset1 = preference_comparisons.TrajectoryDataset(
        cartpole_expert_trajectories,
        rng=np.random.default_rng(0),
    )
    sample1 = dataset1.sample(num_samples)
    dataset2 = preference_comparisons.TrajectoryDataset(
        cartpole_expert_trajectories,
        rng=np.random.default_rng(0),
    )
    sample2 = dataset2.sample(num_samples)

    _check_trajs_equal(sample1, sample2)

    dataset3 = preference_comparisons.TrajectoryDataset(
        cartpole_expert_trajectories,
        rng=np.random.default_rng(42),
    )
    sample3 = dataset3.sample(num_samples)
    with pytest.raises(AssertionError):
        _check_trajs_equal(sample2, sample3)


# CartPole max episode length is 200
@pytest.mark.parametrize("num_steps", [0, 199, 200, 201, 400])
def test_trajectory_dataset_len(
    cartpole_expert_trajectories: Sequence[TrajectoryWithRew],
    num_steps: int,
    rng,
):
    dataset = preference_comparisons.TrajectoryDataset(
        cartpole_expert_trajectories,
        rng=rng,
    )
    sample = dataset.sample(num_steps)
    lengths = [len(t) for t in sample]
    assert sum(lengths) >= num_steps
    if num_steps > 0:
        assert sum(lengths) - min(lengths) < num_steps


def test_trajectory_dataset_too_long(
    cartpole_expert_trajectories: Sequence[TrajectoryWithRew],
    rng,
):
    dataset = preference_comparisons.TrajectoryDataset(
        cartpole_expert_trajectories,
        rng=rng,
    )
    with pytest.raises(RuntimeError, match="Asked for.*but only.* available"):
        dataset.sample(100000)


def test_trajectory_dataset_not_static(
    cartpole_expert_trajectories: Sequence[TrajectoryWithRew],
    rng,
    num_steps: int = 400,
):
    """Tests sample() doesn't always return the same value."""
    dataset = preference_comparisons.TrajectoryDataset(
        cartpole_expert_trajectories,
        rng,
    )
    n_trajectories = len(cartpole_expert_trajectories)
    flakiness_prob = 1 / n_trajectories
    max_flakiness = 1e-6
    # Choose max_samples s.t. flakiness_prob**max_samples <= max_flakiness
    max_samples = math.ceil(math.log(max_flakiness) / math.log(flakiness_prob))
    sample = dataset.sample(num_steps)
    with pytest.raises(AssertionError):
        for _ in range(max_samples):
            sample2 = dataset.sample(num_steps)
            _check_trajs_equal(sample, sample2)


def test_transitions_left_in_buffer(agent_trainer):
    # Faster to just set the counter than to actually fill the buffer
    # with transitions.
    agent_trainer.buffering_wrapper.n_transitions = 2
    with pytest.raises(
        RuntimeError,
        match=re.escape(
            "There are 2 transitions left in the buffer. "
            "Call AgentTrainer.sample() first to clear them.",
        ),
    ):
        agent_trainer.train(steps=1)


@pytest.mark.parametrize(
    "schedule",
    ["constant", "hyperbolic", "inverse_quadratic", lambda t: 1 / (1 + t**3)],
)
def test_preference_comparisons_raises(
    agent_trainer,
    reward_net,
    random_fragmenter,
    preference_model,
    custom_logger,
    schedule,
    rng,
):
    loss = preference_comparisons.CrossEntropyRewardLoss()
    reward_trainer = preference_comparisons.BasicRewardTrainer(
        preference_model,
        loss,
        rng=rng,
    )
    gatherer = preference_comparisons.SyntheticGatherer(rng=rng)
    # no rng, must provide fragmenter, preference gatherer, reward trainer
    no_rng_msg = (
        ".*don't provide.*random state.*provide.*fragmenter"
        ".*preference gatherer.*reward_trainer.*"
    )

    def build_preference_comparsions(gatherer, reward_trainer, fragmenter, rng):
        preference_comparisons.PreferenceComparisons(
            agent_trainer,
            reward_net,
            num_iterations=2,
            transition_oversampling=2,
            reward_trainer=reward_trainer,
            preference_gatherer=gatherer,
            fragmenter=fragmenter,
            custom_logger=custom_logger,
            query_schedule=schedule,
            rng=rng,
        )

    with pytest.raises(ValueError, match=no_rng_msg):
        build_preference_comparsions(gatherer, None, None, rng=None)

    with pytest.raises(ValueError, match=no_rng_msg):
        build_preference_comparsions(None, reward_trainer, None, rng=None)

    with pytest.raises(ValueError, match=no_rng_msg):
        build_preference_comparsions(None, None, random_fragmenter, rng=None)

    # This should not raise
    build_preference_comparsions(gatherer, reward_trainer, random_fragmenter, rng=None)

    # if providing fragmenter, preference gatherer, reward trainer, does not need rng.
    with_rng_msg = (
        "provide.*fragmenter.*preference gatherer.*reward trainer"
        ".*don't need.*random state.*"
    )

    with pytest.raises(ValueError, match=with_rng_msg):
        build_preference_comparsions(
            gatherer,
            reward_trainer,
            random_fragmenter,
            rng=rng,
        )

    # This should not raise
    build_preference_comparsions(None, None, None, rng=rng)
    build_preference_comparsions(gatherer, None, None, rng=rng)
    build_preference_comparsions(None, reward_trainer, None, rng=rng)
    build_preference_comparsions(None, None, random_fragmenter, rng=rng)


@pytest.mark.parametrize(
    "schedule",
    ["constant", "hyperbolic", "inverse_quadratic", lambda t: 1 / (1 + t**3)],
)
def test_trainer_no_crash(
    agent_trainer,
    reward_net,
    random_fragmenter,
    custom_logger,
    schedule,
    rng,
):
    main_trainer = preference_comparisons.PreferenceComparisons(
        agent_trainer,
        reward_net,
        num_iterations=2,
        transition_oversampling=2,
        fragment_length=2,
        fragmenter=random_fragmenter,
        custom_logger=custom_logger,
        query_schedule=schedule,
        initial_epoch_multiplier=2,
        rng=rng,
    )
    result = main_trainer.train(100, 10)
    # We don't expect good performance after training for 10 (!) timesteps,
    # but check stats are within the bounds they should lie in.
    assert result["reward_loss"] > 0.0
    assert 0.0 < result["reward_accuracy"] <= 1.0


def test_reward_ensemble_trainer_raises_type_error(venv, rng):
    reward_net = reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
    preference_model = preference_comparisons.PreferenceModel(
        model=reward_net,
        noise_prob=0.1,
        discount_factor=0.9,
        threshold=50,
    )
    loss = preference_comparisons.CrossEntropyRewardLoss()

    with pytest.raises(
        TypeError,
        match=r"PreferenceModel of a RewardEnsemble expected by EnsembleTrainer.",
    ):
        preference_comparisons.EnsembleTrainer(
            preference_model,
            loss,
            rng=rng,
        )


def test_correct_reward_trainer_used_by_default(
    agent_trainer,
    reward_net,
    random_fragmenter,
    custom_logger,
    rng,
):
    main_trainer = preference_comparisons.PreferenceComparisons(
        agent_trainer,
        reward_net,
        num_iterations=2,
        transition_oversampling=2,
        fragment_length=2,
        fragmenter=random_fragmenter,
        rng=rng,
        custom_logger=custom_logger,
    )

    base_reward_net = reward_net.base if hasattr(reward_net, "base") else reward_net
    if isinstance(base_reward_net, reward_nets.RewardEnsemble):
        assert isinstance(
            main_trainer.reward_trainer,
            preference_comparisons.EnsembleTrainer,
        )
    else:
        assert isinstance(
            main_trainer.reward_trainer,
            preference_comparisons.BasicRewardTrainer,
        )


def test_init_raises_error_when_trying_use_improperly_wrapped_ensemble(
    agent_trainer,
    venv,
    random_fragmenter,
    custom_logger,
    rng,
):
    reward_net = testing_reward_nets.make_ensemble(
        venv.observation_space,
        venv.action_space,
    )
    reward_net = reward_nets.NormalizedRewardNet(reward_net, networks.RunningNorm)
    rgx = (
        r"RewardEnsemble can only be wrapped by "
        r"AddSTDRewardWrapper but found NormalizedRewardNet."
    )
    with pytest.raises(
        ValueError,
        match=rgx,
    ):
        preference_comparisons.PreferenceComparisons(
            agent_trainer,
            reward_net,
            num_iterations=2,
            transition_oversampling=2,
            fragment_length=2,
            fragmenter=random_fragmenter,
            rng=rng,
            custom_logger=custom_logger,
        )


def test_discount_rate_no_crash(
    agent_trainer,
    venv,
    random_fragmenter,
    custom_logger,
    rng,
):
    # also use a non-zero noise probability to check that doesn't cause errors
    reward_net = reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
    preference_model = preference_comparisons.PreferenceModel(
        model=reward_net,
        noise_prob=0.1,
        discount_factor=0.9,
        threshold=50,
    )
    loss = preference_comparisons.CrossEntropyRewardLoss()
    reward_trainer = preference_comparisons.BasicRewardTrainer(
        preference_model,
        loss,
        rng=rng,
    )

    main_trainer = preference_comparisons.PreferenceComparisons(
        agent_trainer,
        reward_net,
        num_iterations=2,
        transition_oversampling=2,
        fragment_length=2,
        fragmenter=random_fragmenter,
        rng=rng,
        reward_trainer=reward_trainer,
        custom_logger=custom_logger,
    )
    main_trainer.train(100, 10)


def create_reward_trainer(
    venv,
    seed: int,
    batch_size: int,
    **kwargs: Any,
):
    th.manual_seed(seed)
    reward_net = reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
    preference_model = preference_comparisons.PreferenceModel(model=reward_net)
    loss = preference_comparisons.CrossEntropyRewardLoss()
    rng = np.random.default_rng(seed)
    reward_trainer = preference_comparisons.BasicRewardTrainer(
        preference_model,
        loss,
        rng=rng,
        batch_size=batch_size,
        **kwargs,
    )
    return reward_trainer, reward_net


def test_gradient_accumulation(
    agent_trainer,
    venv,
    random_fragmenter,
    custom_logger,
    rng,
):
    # Test that training steps on the same dataset with different minibatch sizes
    # result in the same reward network.
    batch_size = 6
    minibatch_size = 3
    num_trajectories = 5

    preference_gatherer = preference_comparisons.SyntheticGatherer(
        custom_logger=custom_logger,
        rng=rng,
    )
    dataset = preference_comparisons.PreferenceDataset()
    trajectory = agent_trainer.sample(num_trajectories)
    fragments = random_fragmenter(trajectory, 1, num_trajectories)
    preferences = preference_gatherer(fragments)
    dataset.push(fragments, preferences)

    seed = rng.integers(2**32)
    reward_trainer1, reward_net1 = create_reward_trainer(venv, seed, batch_size)
    reward_trainer2, reward_net2 = create_reward_trainer(
        venv,
        seed,
        batch_size,
        minibatch_size=minibatch_size,
    )

    for step in range(8):
        print("Step", step)
        seed = rng.integers(2**32)

        th.manual_seed(seed)
        reward_trainer1.train(dataset)

        th.manual_seed(seed)
        reward_trainer2.train(dataset)

        # Note: due to numerical instability, the models are
        # bound to diverge at some point, but should be stable
        # over the short time frame we test over; however, it is
        # theoretically possible that with very unlucky seeding,
        # this could fail.
        atol = 1e-5
        rtol = 1e-4
        for p1, p2 in zip(reward_net1.parameters(), reward_net2.parameters()):
            th.testing.assert_close(p1, p2, atol=atol, rtol=rtol)


def test_synthetic_gatherer_deterministic(
    agent_trainer,
    random_fragmenter,
    rng,
):
    gatherer = preference_comparisons.SyntheticGatherer(
        temperature=0,
        rng=rng,
    )
    trajectories = agent_trainer.sample(10)
    fragments = random_fragmenter(trajectories, fragment_length=2, num_pairs=2)
    preferences1 = gatherer(fragments)
    preferences2 = gatherer(fragments)
    assert np.all(preferences1 == preferences2)


def test_synthetic_gatherer_raises(
    agent_trainer,
    random_fragmenter,
):
    with pytest.raises(
        ValueError,
        match="If `sample` is True, then `rng` must be provided",
    ):
        preference_comparisons.SyntheticGatherer(
            temperature=0,
            sample=True,
        )


def test_fragments_terminal(random_fragmenter):
    trajectories = [
        types.TrajectoryWithRew(
            obs=np.arange(4),
            acts=np.zeros((3,)),
            rews=np.zeros((3,)),
            infos=None,
            terminal=True,
        ),
        types.TrajectoryWithRew(
            obs=np.arange(3),
            acts=np.zeros((2,)),
            rews=np.zeros((2,)),
            infos=None,
            terminal=False,
        ),
    ]
    for _ in range(5):
        for frags in random_fragmenter(trajectories, fragment_length=2, num_pairs=2):
            for frag in frags:
                assert (frag.obs[-1] == 3) == frag.terminal


def test_fragments_too_short_error(agent_trainer):
    trajectories = agent_trainer.sample(2)
    random_fragmenter = preference_comparisons.RandomFragmenter(
        rng=np.random.default_rng(0),
        warning_threshold=0,
    )
    with pytest.raises(
        ValueError,
        match="No trajectories are long enough for the desired fragment length.",
    ):
        # the only important bit is that fragment_length is higher than
        # we'll ever reach
        random_fragmenter(trajectories, fragment_length=10000, num_pairs=2)


def test_preference_dataset_errors(agent_trainer, random_fragmenter):
    dataset = preference_comparisons.PreferenceDataset()
    trajectories = agent_trainer.sample(2)
    fragments = random_fragmenter(trajectories, fragment_length=2, num_pairs=2)
    # just create something with a different shape:
    preferences = np.empty(len(fragments) + 1, dtype=np.float32)
    with pytest.raises(ValueError, match="Unexpected preferences shape"):
        dataset.push(fragments, preferences)

    # Now test dtype
    preferences = np.empty(len(fragments), dtype=np.float64)
    with pytest.raises(ValueError, match="preferences should have dtype float32"):
        dataset.push(fragments, preferences)


def test_preference_dataset_queue(agent_trainer, random_fragmenter, rng):
    dataset = preference_comparisons.PreferenceDataset(max_size=5)
    trajectories = agent_trainer.sample(10)

    gatherer = preference_comparisons.SyntheticGatherer(rng=rng)
    for i in range(6):
        fragments = random_fragmenter(trajectories, fragment_length=2, num_pairs=1)
        preferences = gatherer(fragments)
        assert len(dataset) == min(i, 5)
        dataset.push(fragments, preferences)
        assert len(dataset) == min(i + 1, 5)

    # The first comparison should have been evicted to keep the size at 5
    assert len(dataset) == 5


def test_store_and_load_preference_dataset(
    agent_trainer,
    random_fragmenter,
    tmp_path,
    rng,
):
    dataset = preference_comparisons.PreferenceDataset()
    trajectories = agent_trainer.sample(10)
    fragments = random_fragmenter(trajectories, fragment_length=2, num_pairs=2)
    gatherer = preference_comparisons.SyntheticGatherer(rng=rng)
    preferences = gatherer(fragments)
    dataset.push(fragments, preferences)

    path = tmp_path / "preferences.pkl"
    dataset.save(path)
    loaded = preference_comparisons.PreferenceDataset.load(path)
    assert len(loaded) == len(dataset)
    for sample, loaded_sample in zip(dataset, loaded):
        fragments, preference = sample
        loaded_fragments, loaded_preference = loaded_sample

        assert preference == loaded_preference
        _check_trajs_equal(fragments, loaded_fragments)


def test_exploration_no_crash(
    agent,
    reward_net,
    venv,
    random_fragmenter,
    custom_logger,
    rng,
):
    agent_trainer = preference_comparisons.AgentTrainer(
        agent,
        reward_net,
        venv,
        rng=rng,
        exploration_frac=0.5,
    )
    main_trainer = preference_comparisons.PreferenceComparisons(
        agent_trainer,
        reward_net,
        num_iterations=2,
        transition_oversampling=2,
        fragment_length=5,
        fragmenter=random_fragmenter,
        rng=rng,
        custom_logger=custom_logger,
    )
    main_trainer.train(100, 10)


@pytest.mark.parametrize("uncertainty_on", UNCERTAINTY_ON)
def test_active_fragmenter_discount_rate_no_crash(
    agent_trainer,
    venv,
    random_fragmenter,
    uncertainty_on,
    custom_logger,
    rng,
):
    # also use a non-zero noise probability to check that doesn't cause errors
    reward_net = reward_nets.RewardEnsemble(
        venv.observation_space,
        venv.action_space,
        members=[
            reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
            for _ in range(2)
        ],
    )
    preference_model = preference_comparisons.PreferenceModel(
        model=reward_net,
        noise_prob=0.1,
        discount_factor=0.9,
        threshold=50,
    )

    fragmenter = preference_comparisons.ActiveSelectionFragmenter(
        preference_model=preference_model,
        base_fragmenter=random_fragmenter,
        fragment_sample_factor=2,
        uncertainty_on=uncertainty_on,
        custom_logger=custom_logger,
    )

    preference_model = preference_comparisons.PreferenceModel(
        model=reward_net,
        noise_prob=0.1,
        discount_factor=0.9,
        threshold=50,
    )
    loss = preference_comparisons.CrossEntropyRewardLoss()

    reward_trainer = preference_comparisons.EnsembleTrainer(
        preference_model,
        loss,
        rng=rng,
    )

    main_trainer = preference_comparisons.PreferenceComparisons(
        agent_trainer,
        reward_net,
        num_iterations=2,
        transition_oversampling=2,
        fragment_length=2,
        fragmenter=fragmenter,
        rng=rng,
        reward_trainer=reward_trainer,
        custom_logger=custom_logger,
    )
    main_trainer.train(100, 10)


@pytest.fixture(scope="module")
def interval_param_scaler() -> updaters.IntervalParamScaler:
    return updaters.IntervalParamScaler(
        scaling_factor=0.1,
        tolerable_interval=(1.1, 1.5),
    )


def test_reward_trainer_regularization_no_crash(
    agent_trainer,
    venv,
    random_fragmenter,
    custom_logger,
    preference_model,
    interval_param_scaler,
    rng,
):
    reward_net = reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
    loss = preference_comparisons.CrossEntropyRewardLoss()
    initial_lambda = 0.1
    regularizer_factory = regularizers.LpRegularizer.create(
        initial_lambda=initial_lambda,
        val_split=0.2,
        lambda_updater=interval_param_scaler,
        p=2,
    )
    reward_trainer = preference_comparisons.BasicRewardTrainer(
        preference_model,
        loss,
        regularizer_factory=regularizer_factory,
        custom_logger=custom_logger,
        rng=rng,
    )

    main_trainer = preference_comparisons.PreferenceComparisons(
        agent_trainer,
        reward_net,
        num_iterations=2,
        transition_oversampling=2,
        fragment_length=2,
        fragmenter=random_fragmenter,
        reward_trainer=reward_trainer,
        custom_logger=custom_logger,
        rng=rng,
    )
    main_trainer.train(50, 50)


def test_reward_trainer_regularization_raises(
    agent_trainer,
    venv,
    random_fragmenter,
    custom_logger,
    preference_model,
    interval_param_scaler,
    rng,
):
    reward_net = reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
    loss = preference_comparisons.CrossEntropyRewardLoss()
    initial_lambda = 0.1
    regularizer_factory = regularizers.LpRegularizer.create(
        initial_lambda=initial_lambda,
        val_split=0.2,
        lambda_updater=interval_param_scaler,
        p=2,
    )
    reward_trainer = preference_comparisons.BasicRewardTrainer(
        preference_model,
        loss,
        regularizer_factory=regularizer_factory,
        custom_logger=custom_logger,
        rng=rng,
    )

    main_trainer = preference_comparisons.PreferenceComparisons(
        agent_trainer,
        reward_net,
        num_iterations=2,
        transition_oversampling=2,
        fragment_length=2,
        fragmenter=random_fragmenter,
        reward_trainer=reward_trainer,
        custom_logger=custom_logger,
        rng=rng,
    )
    with pytest.raises(
        ValueError,
        match="Not enough data samples to split " "into training and validation.*",
    ):
        main_trainer.train(100, 10)


@pytest.fixture
def ensemble_preference_model(venv) -> preference_comparisons.PreferenceModel:
    reward_net = reward_nets.RewardEnsemble(
        venv.observation_space,
        venv.action_space,
        members=[
            reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
            for _ in range(2)
        ],
    )
    return preference_comparisons.PreferenceModel(
        model=reward_net,
        noise_prob=0.1,
        discount_factor=0.9,
        threshold=50,
    )


@pytest.fixture
def preference_model(venv) -> preference_comparisons.PreferenceModel:
    reward_net = reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
    return preference_comparisons.PreferenceModel(
        model=reward_net,
        noise_prob=0.1,
        discount_factor=0.9,
        threshold=50,
    )


def test_active_fragmenter_uncertainty_on_not_supported_error(
    ensemble_preference_model,
    random_fragmenter,
):
    re_match = r".* not supported\.\n\s+`uncertainty_on` should be from .*"
    with pytest.raises(ValueError, match=re_match):
        preference_comparisons.ActiveSelectionFragmenter(
            preference_model=ensemble_preference_model,
            base_fragmenter=random_fragmenter,
            fragment_sample_factor=2,
            uncertainty_on="uncertainty_on",
        )

    with pytest.raises(ValueError, match=re_match):
        fragmenter = preference_comparisons.ActiveSelectionFragmenter(
            preference_model=ensemble_preference_model,
            base_fragmenter=random_fragmenter,
            fragment_sample_factor=2,
            uncertainty_on="logit",
        )
        fragmenter._uncertainty_on = "uncertainty_on"
        members = ensemble_preference_model.model.num_members
        fragmenter.variance_estimate(th.rand(10, members), th.rand(10, members))


def test_active_selection_raises_error_when_initialized_without_an_ensemble(
    preference_model,
    random_fragmenter,
):
    with pytest.raises(
        ValueError,
        match=r"PreferenceModel not wrapped over an ensemble.*",
    ):
        preference_comparisons.ActiveSelectionFragmenter(
            preference_model=preference_model,
            base_fragmenter=random_fragmenter,
            fragment_sample_factor=2,
            uncertainty_on="logit",
        )


def test_agent_trainer_populates_buffer(agent_trainer):
    agent_trainer.train(steps=1)
    assert agent_trainer.buffering_wrapper.n_transitions > 0


def test_agent_trainer_sample(venv, agent_trainer):
    trajectories = agent_trainer.sample(2)
    assert len(trajectories) > 0
    assert all(
        trajectory.obs.shape[1:] == venv.observation_space.shape
        for trajectory in trajectories
    )


def test_agent_trainer_sample_image_observations(rng):
    """Test `AgentTrainer.sample()` in an image environment.

    SB3 algorithms may rearrange the channel dimension in environments with image
    observations, but `sample()` should return observations matching the original
    environment.

    Args:
        rng: Random number generator (with a fixed seed).
    """
    venv = DummyVecEnv([lambda: FakeImageEnv()])
    reward_net = reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
    agent = stable_baselines3.PPO(
        "MlpPolicy",
        venv,
        n_epochs=1,
        batch_size=2,
        n_steps=10,
    )
    agent_trainer = preference_comparisons.AgentTrainer(
        agent,
        reward_net,
        venv,
        exploration_frac=0.5,
        rng=rng,
    )
    trajectories = agent_trainer.sample(2)
    assert len(trajectories) > 0
    assert all(
        trajectory.obs.shape[1:] == venv.observation_space.shape
        for trajectory in trajectories
    )


class ActionIsRewardEnv(gym.Env):
    """Two step environment where the reward is the sum of actions."""

    def __init__(self):
        """Initialize environment."""
        super().__init__()
        self.action_space = spaces.Discrete(50)
        self.observation_space = gym.spaces.Box(np.array([0]), np.array([1]))
        self.steps = 0

    def step(self, action):
        obs = np.array([0])
        reward = action
        # Some algorithms expect trajectories at least two steps long,
        # so we allow two steps to be taken.
        done = self.steps > 0
        info = {}
        self.steps += 1
        return obs, reward, done, False, info

    def reset(self, seed=None):
        self.steps = 0.0
        return np.array([0]), {}


@pytest.fixture
def action_is_reward_venv():
    return DummyVecEnv(
        [ActionIsRewardEnv],
    )


@pytest.fixture
def action_is_reward_agent(action_is_reward_venv, rng):
    return stable_baselines3.PPO(
        "MlpPolicy",
        action_is_reward_venv,
        n_epochs=1,
        batch_size=2,
        n_steps=10,
    )


def basic_reward_trainer(venv, rng):
    loss = preference_comparisons.CrossEntropyRewardLoss()
    reward_net = reward_nets.BasicRewardNet(
        venv.observation_space,
        venv.action_space,
    )
    preference_model = preference_comparisons.PreferenceModel(
        model=reward_net,
        noise_prob=0.1,
        discount_factor=0.9,
        threshold=50,
    )
    return preference_comparisons.BasicRewardTrainer(
        preference_model,
        loss,
        rng=rng,
        lr=1e-4,
    )


def ensemble_reward_trainer(venv, rng):
    loss = preference_comparisons.CrossEntropyRewardLoss()
    reward_net = reward_nets.RewardEnsemble(
        venv.observation_space,
        venv.action_space,
        members=[
            reward_nets.BasicRewardNet(
                venv.observation_space,
                venv.action_space,
            )
            for _ in range(3)
        ],
    )
    preference_model = preference_comparisons.PreferenceModel(
        model=reward_net,
        noise_prob=0.1,
        discount_factor=0.9,
        threshold=50,
    )
    return preference_comparisons.EnsembleTrainer(
        preference_model,
        loss,
        rng=rng,
        lr=1e-4,
    )


@pytest.mark.parametrize(
    "action_is_reward_trainer_func",
    [basic_reward_trainer, ensemble_reward_trainer],
)
def test_that_trainer_improves(
    action_is_reward_venv,
    action_is_reward_agent,
    action_is_reward_trainer_func,
    random_fragmenter,
    custom_logger,
    rng,
):
    """Tests that training improves performance of the reward network and agent."""
    action_is_reward_trainer = action_is_reward_trainer_func(action_is_reward_venv, rng)
    agent_trainer = preference_comparisons.AgentTrainer(
        action_is_reward_agent,
        action_is_reward_trainer._preference_model.model,
        action_is_reward_venv,
        rng,
    )

    main_trainer = preference_comparisons.PreferenceComparisons(
        agent_trainer,
        action_is_reward_trainer._preference_model.model,
        num_iterations=2,
        transition_oversampling=2,
        fragment_length=2,
        fragmenter=random_fragmenter,
        rng=rng,
        reward_trainer=action_is_reward_trainer,
        custom_logger=custom_logger,
    )

    # Get initial agent performance
    novice_agent_rewards, _ = evaluation.evaluate_policy(
        agent_trainer.algorithm.policy,
        action_is_reward_venv,
        50,
        return_episode_rewards=True,
    )

    # Train for a short period of time, and then again.
    # We expect the reward network to have a better estimate of the reward
    # after this training, and thus `later_rewards` should have lower loss.
    first_reward_network_stats = main_trainer.train(20, 20)

    later_reward_network_stats = main_trainer.train(100, 40)
    assert (
        first_reward_network_stats["reward_loss"]
        > later_reward_network_stats["reward_loss"]
    )

    # The agent should have also improved
    trained_agent_rewards, _ = evaluation.evaluate_policy(
        agent_trainer.algorithm.policy,
        action_is_reward_venv,
        50,
        return_episode_rewards=True,
    )

    assert reward_improvement.is_significant_reward_improvement(
        novice_agent_rewards,
        trained_agent_rewards,
    )



"""Load serialized reward functions of different types."""

from typing import Any, Callable, Dict, Iterable, Optional, Sequence, Type, Union, cast

import numpy as np
import torch as th
from stable_baselines3.common.vec_env import VecEnv

from imitation.rewards import reward_function, reward_nets
from imitation.util import registry, util

# TODO(sam): I suspect this whole file can be replaced with th.load calls. Try
# that refactoring once I have things running.

RewardFnLoaderFn = Callable[[str, VecEnv], reward_function.RewardFn]

reward_registry: registry.Registry[RewardFnLoaderFn] = registry.Registry()


class ValidateRewardFn(reward_function.RewardFn):
    """Wrap reward function to add sanity check.

    Checks that the length of the reward vector is equal to the batch size of the input.
    """

    def __init__(
        self,
        reward_fn: reward_function.RewardFn,
    ) -> None:
        """Builds the reward validator.

        Args:
            reward_fn: base reward function
        """
        super().__init__()
        self.reward_fn = reward_fn

    def __call__(
        self,
        state: np.ndarray,
        action: np.ndarray,
        next_state: np.ndarray,
        done: np.ndarray,
    ) -> np.ndarray:
        rew = self.reward_fn(state, action, next_state, done)
        assert rew.shape == (len(state),)
        return rew


def _strip_wrappers(
    reward_net: reward_nets.RewardNet,
    wrapper_types: Iterable[Type[reward_nets.RewardNetWrapper]],
) -> reward_nets.RewardNet:
    """Attempts to remove provided wrappers.

    Strips wrappers of type `wrapper_type` from `reward_net` in order until either the
    wrapper type to remove does not match the type of net or there are no more wrappers
    to remove.

    Args:
        reward_net: an instance of a reward network that may be wrapped
        wrapper_types: an iterable of wrapper types in the order they should be removed

    Returns:
        The reward network with the listed wrappers removed
    """
    for wrapper_type in wrapper_types:
        assert issubclass(
            wrapper_type,
            reward_nets.RewardNetWrapper,
        ), f"trying to remove non-wrapper type {wrapper_type}"

        if isinstance(reward_net, wrapper_type):
            reward_net = reward_net.base
        else:
            break

    return reward_net


def _make_functional(
    net: reward_nets.RewardNet,
    attr: str = "predict",
    default_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> reward_function.RewardFn:
    if default_kwargs is None:
        default_kwargs = {}
    default_kwargs.update(kwargs)
    return lambda *args: getattr(net, attr)(*args, **default_kwargs)


WrapperPrefix = Sequence[Type[reward_nets.RewardNet]]


def _prefix_matches(wrappers: Sequence[Type[Any]], prefix: Sequence[Type[Any]]) -> bool:
    """Return true if `prefix` is a prefix of `wrappers`."""
    # Base cases
    if len(prefix) == 0:
        # If we run out of prefix before running out of wrappers
        return True
    elif len(wrappers) == 0:
        # If we run out of wrappers before we run out of prefix
        return False

    prefix_head, *prefix_tail = prefix
    wrappers_head, *wrappers_tail = wrappers

    if not issubclass(wrappers_head, prefix_head):
        return False

    return _prefix_matches(wrappers_tail, prefix_tail)


def _validate_wrapper_structure(
    reward_net: Union[reward_nets.RewardNet, reward_nets.RewardNetWrapper],
    prefixes: Iterable[WrapperPrefix],
) -> reward_nets.RewardNet:
    """Reward net if it has a valid structure.

    A wrapper prefix specifies, from outermost to innermost, which wrappers must
    be present. If any of the wrapper prefixes match then the RewardNet is considered
    valid.

    Args:
        reward_net: net to test
        prefixes: A list of acceptable wrapper prefixes.

    Returns:
        the reward_net if it is valid

    Raises:
        TypeError: if the wrapper structure is not valid with a useful message.

    >>> class RewardNetA(RewardNet):
    ...     def forward(*args):
    ...         pass
    >>> class WrapperB(RewardNetWrapper):
    ...     def forward(*args):
    ...         pass
    >>> reward_net = RewardNetA(None, None)
    >>> reward_net = WrapperB(reward_net)
    >>> assert isinstance(reward_net.base, RewardNet)
    >>> reward_net == _validate_wrapper_structure(reward_net, [[WrapperB, RewardNetA]]))
    True
    """
    wrapper = reward_net
    wrappers = []
    while hasattr(wrapper, "base"):
        wrappers.append(wrapper.__class__)
        wrapper = cast(reward_nets.RewardNet, wrapper.base)
    wrappers.append(wrapper.__class__)  # append the final reward net

    if any(_prefix_matches(wrappers, prefix) for prefix in prefixes):
        return reward_net

    # Otherwise provide a useful error
    formatted_prefixes = [
        "[" + ",".join(t.__name__ for t in prefix) + "]" for prefix in prefixes
    ]

    formatted_wrapper_structure = "[" + ",".join(t.__name__ for t in wrappers) + "]"

    raise TypeError(
        "Wrapper structure should"
        + " match "
        + " or ".join(formatted_prefixes)
        + " but found "
        + formatted_wrapper_structure,
    )


def load_zero(path: str, venv: VecEnv) -> reward_function.RewardFn:
    del path, venv

    def f(
        state: np.ndarray,
        action: np.ndarray,
        next_state: np.ndarray,
        done: np.ndarray,
    ) -> np.ndarray:
        del action, next_state, done  # Unused.
        return np.zeros(state.shape[0])

    return f


# TODO(adam): I think we can get rid of this and have just one RewardNet.

reward_registry.register(
    key="RewardNet_shaped",
    value=lambda path, _, **kwargs: ValidateRewardFn(
        _make_functional(
            _validate_wrapper_structure(
                th.load(str(path)),
                {(reward_nets.ShapedRewardNet,)},
            ),
        ),
    ),
)

reward_registry.register(
    key="RewardNet_unshaped",
    value=lambda path, _, **kwargs: ValidateRewardFn(
        _make_functional(
            _strip_wrappers(th.load(str(path)), (reward_nets.ShapedRewardNet,)),
        ),
    ),
)

reward_registry.register(
    key="RewardNet_normalized",
    value=lambda path, _, **kwargs: ValidateRewardFn(
        _make_functional(
            _validate_wrapper_structure(
                th.load(str(path)),
                {(reward_nets.NormalizedRewardNet,)},
            ),
            attr="predict_processed",
            default_kwargs={"update_stats": False},
            **kwargs,
        ),
    ),
)

reward_registry.register(
    key="RewardNet_unnormalized",
    value=lambda path, _, **kwargs: ValidateRewardFn(
        _make_functional(
            _strip_wrappers(th.load(str(path)), (reward_nets.NormalizedRewardNet,)),
        ),
    ),
)

reward_registry.register(
    key="RewardNet_std_added",
    value=lambda path, _, **kwargs: ValidateRewardFn(
        _make_functional(
            _strip_wrappers(
                _validate_wrapper_structure(
                    th.load(str(path)),
                    {
                        (reward_nets.AddSTDRewardWrapper,),
                        (
                            reward_nets.NormalizedRewardNet,
                            reward_nets.AddSTDRewardWrapper,
                        ),
                    },
                ),
                (reward_nets.NormalizedRewardNet,),
            ),
            attr="predict_processed",
            default_kwargs={},
            **kwargs,
        ),
    ),
)

reward_registry.register(key="zero", value=load_zero)


@util.docstring_parameter(reward_types=", ".join(reward_registry.keys()))
def load_reward(
    reward_type: str,
    reward_path: str,
    venv: VecEnv,
    **kwargs: Any,
) -> reward_function.RewardFn:
    """Load serialized reward.

    Args:
        reward_type: A key in `reward_registry`. Valid types
            include {reward_types}.
        reward_path: A path specifying the reward.
        venv: An environment that the policy is to be used with.
        **kwargs: kwargs to pass to reward fn

    Returns:
        The deserialized reward.
    """
    reward_loader = reward_registry.get(reward_type)
    return reward_loader(reward_path, venv, **kwargs)



"""Configuration settings for eval_policy, evaluating pre-trained policies."""

import sacred

from imitation.scripts.ingredients import environment, expert
from imitation.scripts.ingredients import logging as logging_ingredient

eval_policy_ex = sacred.Experiment(
    "eval_policy",
    ingredients=[
        logging_ingredient.logging_ingredient,
        environment.environment_ingredient,
        expert.expert_ingredient,
    ],
)


@eval_policy_ex.config
def replay_defaults():
    eval_n_timesteps = int(1e4)  # Min timesteps to evaluate, optional.
    eval_n_episodes = None  # Num episodes to evaluate, optional.

    videos = False  # save video files
    video_kwargs = {}  # arguments to VideoWrapper
    render = False  # render to screen
    render_fps = 60  # -1 to render at full speed

    reward_type = None  # Optional: override with reward of this type
    reward_path = None  # Path of serialized reward to load

    rollout_save_path = None  # where to save rollouts to -- if None, do not save

    explore_kwargs = (
        None  # kwargs to feed to ExplorationWrapper -- if None, do not wrap
    )


@eval_policy_ex.named_config
def explore_eps_greedy():
    explore_kwargs = dict(switch_prob=1.0, random_prob=0.1)


@eval_policy_ex.named_config
def render():
    environment = dict(num_vec=1, parallel=False)
    render = True


@eval_policy_ex.named_config
def acrobot():
    environment = dict(gym_id="Acrobot-v1")


@eval_policy_ex.named_config
def ant():
    environment = dict(gym_id="Ant-v2")


@eval_policy_ex.named_config
def cartpole():
    environment = dict(gym_id="CartPole-v1")


@eval_policy_ex.named_config
def seals_cartpole():
    environment = dict(gym_id="seals/CartPole-v0")


@eval_policy_ex.named_config
def half_cheetah():
    environment = dict(gym_id="HalfCheetah-v2")


@eval_policy_ex.named_config
def seals_half_cheetah():
    environment = dict(gym_id="seals/HalfCheetah-v0")


@eval_policy_ex.named_config
def seals_hopper():
    environment = dict(gym_id="seals/Hopper-v0")


@eval_policy_ex.named_config
def seals_humanoid():
    environment = dict(gym_id="seals/Humanoid-v0")


@eval_policy_ex.named_config
def mountain_car():
    environment = dict(gym_id="MountainCar-v0")


@eval_policy_ex.named_config
def seals_mountain_car():
    environment = dict(gym_id="seals/MountainCar-v0")


@eval_policy_ex.named_config
def pendulum():
    environment = dict(gym_id="Pendulum-v1")


@eval_policy_ex.named_config
def reacher():
    environment = dict(gym_id="Reacher-v2")


@eval_policy_ex.named_config
def seals_ant():
    environment = dict(gym_id="seals/Ant-v0")


@eval_policy_ex.named_config
def seals_swimmer():
    environment = dict(gym_id="seals/Swimmer-v0")


@eval_policy_ex.named_config
def seals_walker():
    environment = dict(gym_id="seals/Walker2d-v0")


@eval_policy_ex.named_config
def fast():
    environment = dict(gym_id="seals/CartPole-v0", num_vec=1, parallel=False)
    render = True
    eval_n_timesteps = 1
    eval_n_episodes = None



"""This is a simple example demonstrating how to clone the behavior of an expert.

Refer to the jupyter notebooks for more detailed examples of how to use the algorithms.
"""
import numpy as np
from stable_baselines3 import PPO
from stable_baselines3.common.evaluation import evaluate_policy
from stable_baselines3.ppo import MlpPolicy

from imitation.algorithms import bc
from imitation.data import rollout
from imitation.data.wrappers import RolloutInfoWrapper
from imitation.policies.serialize import load_policy
from imitation.util.util import make_vec_env

rng = np.random.default_rng(0)
env = make_vec_env(
    "seals:seals/CartPole-v0",
    rng=rng,
    post_wrappers=[lambda env, _: RolloutInfoWrapper(env)],  # for computing rollouts
)


def train_expert():
    # note: use `download_expert` instead to download a pretrained, competent expert
    print("Training a expert.")
    expert = PPO(
        policy=MlpPolicy,
        env=env,
        seed=0,
        batch_size=64,
        ent_coef=0.0,
        learning_rate=0.0003,
        n_epochs=10,
        n_steps=64,
    )
    expert.learn(1_000)  # Note: change this to 100_000 to train a decent expert.
    return expert


def download_expert():
    print("Downloading a pretrained expert.")
    expert = load_policy(
        "ppo-huggingface",
        organization="HumanCompatibleAI",
        env_name="seals-CartPole-v0",
        venv=env,
    )
    return expert


def sample_expert_transitions():
    # expert = train_expert()  # uncomment to train your own expert
    expert = download_expert()

    print("Sampling expert transitions.")
    rollouts = rollout.rollout(
        expert,
        env,
        rollout.make_sample_until(min_timesteps=None, min_episodes=50),
        rng=rng,
    )
    return rollout.flatten_trajectories(rollouts)


transitions = sample_expert_transitions()
bc_trainer = bc.BC(
    observation_space=env.observation_space,
    action_space=env.action_space,
    demonstrations=transitions,
    rng=rng,
)

evaluation_env = make_vec_env(
    "seals:seals/CartPole-v0",
    rng=rng,
    env_make_kwargs={"render_mode": "human"},  # for rendering
)

print("Evaluating the untrained policy.")
reward, _ = evaluate_policy(
    bc_trainer.policy,  # type: ignore[arg-type]
    evaluation_env,
    n_eval_episodes=3,
    render=True,  # comment out to speed up
)
print(f"Reward before training: {reward}")

print("Training a policy using Behavior Cloning")
bc_trainer.train(n_epochs=1)

print("Evaluating the trained policy.")
reward, _ = evaluate_policy(
    bc_trainer.policy,  # type: ignore[arg-type]
    evaluation_env,
    n_eval_episodes=3,
    render=True,  # comment out to speed up
)
print(f"Reward after training: {reward}")
